/* -*- mode: C; c-file-style: "Stroustrup" -*-
 *
 * das08jr
 * =======
 *
 * Linux device driver for Computer Boards CIO-DAS08/Jr-AO board.
 *
 * This board has 8 analog input lines and 2 analog output lines. Additionally
 * there are 8 digital outputs.
 * These lines are mapped onto minor 0 (digital output), 1 - 8 for analog
 * input, and 1 + 2 for analog output by this driver.
 *
 * Copyright (C) 1998 Jochen Kuepper
 *
 * $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */


#define __KERNEL__
#define MODULE


#include <linux/config.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/system.h>

#include "das08jr.h"



/* port addresses to autoprobe
 * If the parameter io=0xNNN is not given at load time, these addresses are
 * probed in their respective order until one device is found or all specified
 * addresses are probed */
static unsigned long autoprobe_port[] =
{
	0x200,
	0x270,
	0x2e0,
	0x2e8,
	0x2f0,
	0x2f8,
	0x300,
	0x310,
	0x320,
	0x378,
	0
};



/* driver version */
static const char das08jr_version[] = "$Revision: 1.3 $";



/* user configurable port base
 * If you give a parameter io=0xNNN, this address will be used,
 * autoprobing will be done otherwise. */
unsigned long io = 0;



/* number of available lines */
#define DAS08JR_ANALOG_INPUTS    ( 8 )
#define DAS08JR_ANALOG_OUTPUTS   ( 2 )
#define DAS08JR_DIGITAL_LINES    ( 1 )
#define DAS08JR_MAX_LINE         ( 8 ) /* we have nine lines starting at zero */

/* number of ports */
#define DAS08JR_SIZE             ( 8 )

/* Port Offsets */
#define DAS08JR_ANALOG_IN_LSB    ( 0 ) /* read */
#define DAS08JR_ANALOG_IN_MSB    ( 1 ) /* read */
#define DAS08JR_START            ( 1 )
#define DAS08JR_STATUS           ( 2 ) /* read */
#define DAS08JR_MUX              ( 2 ) /* read */
#define DAS08JR_CONTROL          ( 2 )
#define DAS08JR_DIGITAL_IN       ( 3 ) /* read */
#define DAS08JR_DIGITAL_OUT      ( 3 )
#define DAS08JR_ANALOG_OUT_LSB0  ( 4 )
#define DAS08JR_ANALOG_OUT_MSB0  ( 5 )
#define DAS08JR_ANALOG_OUT_LSB1  ( 6 )
#define DAS08JR_ANALOG_OUT_MSB1  ( 7 )



/* local functions */
static unsigned long    das08jr_check_port( unsigned long );
static int              das08jr_open( struct inode *, struct file * );
static void             das08jr_release( struct inode *, struct file * );
static int              das08jr_read( struct inode *, struct file *, char *, int );
static int              das08jr_read_digital( struct inode *, struct file *, char *, int );
static int              das08jr_write_digital( struct inode *, struct file *, const char *, int );
static int              das08jr_write_ro( struct inode *, struct file *, const char *, int );
static int              das08jr_write_rw( struct inode *, struct file *, const char *, int );



/* implemented file operations for digital lines */
static struct file_operations das08jr_fops_digital =
{
    NULL,
    das08jr_read_digital,
    das08jr_write_digital,
    NULL,
    NULL,
    NULL,
    NULL,
    das08jr_open,
    das08jr_release,
    /* no more */
};
    
/* implemented file operations for readonly analog lines ( 3 - 8 ) */
static struct file_operations das08jr_fops_ro =
{
    NULL,
    das08jr_read,
    das08jr_write_ro,
    NULL,
    NULL,
    NULL,
    NULL,
    das08jr_open,
    das08jr_release,
    /* no more */
};
    
/* implemented file operations for writable analog lines( 1 + 2 ) */
static struct file_operations das08jr_fops_rw =
{
    NULL,
    das08jr_read,
    das08jr_write_rw,
    NULL,
    NULL,
    NULL,
    NULL,
    das08jr_open,
    das08jr_release,
    /* no more */
};



/**
 * Probe for device, register ports, register as character device,
 * export exactly the symbols we want to be public (none yet).
 *
 * @return Success status, zero on success.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
int init_module( void )
{
	/* check device port base and request the port region */
    if( 0 == io ) {
		int i = 0;
		while ( 0 != ( io = autoprobe_port[ i ] ) ) {
			if( 0 != das08jr_check_port( io ) )
				break;
			i++;
		}
		if( 0 == io ) {
			return( -ENODEV );
		}
    } else {
		if( 0 == das08jr_check_port( io ) )
			return( -ENODEV );
	}
	/* register as character device */
	if( 0 > register_chrdev( DAS08JR_MAJOR, DAS08JR_NAME, &das08jr_fops_rw ) ) {
		printk( KERN_WARNING DAS08JR_NAME ": Cannot register major number %d.\n", DAS08JR_MAJOR );
		release_region( io, DAS08JR_SIZE );
		return( -EBUSY );
	} 
	/* hide all symbols */
	register_symtab( NULL );
	/* finish up */
	printk( KERN_INFO DAS08JR_NAME ": CIO-DAS08Jr-AO driver (%s) loaded\n"
			DAS08JR_NAME ": Copyright (C) 1998 Jochen Kuepper\n"
			DAS08JR_NAME ": found board at 0x%lx\n",
			das08jr_version, io );
    return( 0 );
}



/**
 * Cleanup when we are unloaded.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */ 
void cleanup_module( void )
{
    unregister_chrdev( DAS08JR_MAJOR, DAS08JR_NAME );
    release_region( io, DAS08JR_SIZE );
    printk( KERN_INFO DAS08JR_NAME " unloaded\n" );
    return;
}



/**
 * Check the given port base address.
 *
 * @param base Port address to check for a device at.
 * @return The port base if we found a conforming device, 0 otherwise.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */ 
static unsigned long das08jr_check_port( unsigned long base )
{
	/* check and register given port region */
	if( -EBUSY == check_region( base, DAS08JR_SIZE ) )
		return 0;
	request_region( base, DAS08JR_SIZE, DAS08JR_NAME );
	/* Try out the port */
	outb_p( 0x03, base + DAS08JR_CONTROL );
	if( 0x03 != ( 0x07 & inb_p( base + DAS08JR_MUX ) ) ) {
		printk( KERN_DEBUG DAS08JR_NAME ": no device at %lx\n", base );
		release_region( base, DAS08JR_SIZE );
		return 0;
	}
	outb_p( 0x06, base + DAS08JR_CONTROL );
	if( 0x06 != ( 0x07 & inb_p( base + DAS08JR_MUX ) ) ) {
		printk( KERN_DEBUG DAS08JR_NAME ": no device at %lx\n", base );
		release_region( base, DAS08JR_SIZE );
		return 0;
	}
	return( base );
}



/**
 * Open device.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */ 
static int das08jr_open( struct inode *inode, struct file *fp )
{
    int minor = MINOR( inode->i_rdev );
    if( DAS08JR_MAX_LINE < minor )
	return( -ENODEV );
    if( DAS08JR_DIGITAL_LINES > minor ) {
	/* digital lines */
	fp->f_op = &das08jr_fops_digital;
    } else if( DAS08JR_ANALOG_OUTPUTS + DAS08JR_DIGITAL_LINES > minor ) {
	/* readonly analog lines */
	if( FMODE_WRITE & fp->f_mode )
	    return( -EACCES );
	fp->f_op = &das08jr_fops_ro;
    }
    MOD_INC_USE_COUNT;
    return( 0 );
}



/**
 * Close device
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */ 
static void das08jr_release( struct inode *inode, struct file *fp )
{
    MOD_DEC_USE_COUNT;
    return;
}



/**
 * Read digital lines
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
static int das08jr_read_digital( struct inode *inode, struct file *fp, char *buf, int count )
{
    unsigned char in = inb( io + DAS08JR_DIGITAL_IN );
    put_user( in, buf );
    return( 1 );
}



/**
 * write digital lines
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
static int das08jr_write_digital( struct inode *inode, struct file *fp, const char *buf, int count )
{
    unsigned char out = get_user( buf );
    outb( out, io + DAS08JR_DIGITAL_OUT );
    return( 1 );
}



/**
 * Read from analog line
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
static int das08jr_read( struct inode *inode, struct file *fp, char *buf, int count )
{
    /* we alway read one short value */
    if( count >= 2 ) {
	/* start daq */
	/* busy waiting for board to finish */
	/* udelay( DAS08JR_DELAY ); */
	/* read data and put it into user space */
	unsigned char lsb = inb( io + DAS08JR_ANALOG_IN_LSB );
	unsigned char msb = inb( io + DAS08JR_ANALOG_IN_MSB );
	unsigned short val = ( ( (int)msb ) << 4 ) + ( lsb >> 4 );
	put_user( val, (unsigned short *)buf );
	return( 2 );
    }
    return( 0 );
}



/**
 * Write to writeable analog line.
 *
 * We do not need to check wether this really is an writeable line, since open
 * cares about that an we only get called on lines 1 or 2.
 *
 * We simply put one value to the output ports, start D/A conversion, and return.
 *
 * @return On success the number of written bytes ( always 2 ),
 *         a negative error code otherwise.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
static int das08jr_write_rw( struct inode *inode, struct file *fp, const char *buf, int count )
{
    /* we need a short value */
    if( count >= 2 ) {
	unsigned short val = get_user( (unsigned short *)buf );
	if( DAS08JR_ANALOG_OUTPUTS == MINOR( inode->i_rdev ) ) {
	    outb( (unsigned char)val >> 4, DAS08JR_ANALOG_OUT_MSB1 );
	    outb( (unsigned char)val << 4, DAS08JR_ANALOG_OUT_LSB1 );
	} else {
	    outb( (unsigned char)val >> 4, DAS08JR_ANALOG_OUT_MSB0 );
	    outb( (unsigned char)val << 4, DAS08JR_ANALOG_OUT_LSB0 );
	}
	outb( 0xff, DAS08JR_START );
    }
    return( 0 );
}



/**
 * Write to readonly analog line.
 *
 * @return Since we are always called on readonly lines, simply return -EPERM.
 *
 * @author Jochen Kpper
 * @version $Id: das08jr.c,v 1.3 1998/11/09 18:57:05 jochen Exp $
 */
static int das08jr_write_ro( struct inode *inode, struct file *fp, const char *buf, int count )
{
    return( -EPERM );
}
